/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;
import com.bitctrl.util.Timestamp;

import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.FuzzyWert;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.OdFuzzyWert;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateEvent;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateListener;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.MesswertDatum;
import de.bsvrz.sys.funclib.bitctrl.modell.OnlineDatensatz;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;

/**
 * Basisklasse für einen Testfall der Straßensubsegmentanalyse.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 * @param <T>
 *            der Typ des Messwertdatums.
 */
@SuppressWarnings("nls")
public abstract class Testfall<T extends MesswertDatum> {

	/**
	 * Empänger für die Istwerte.
	 */
	private class Empfaenger implements DatensatzUpdateListener {

		@Override
		public void datensatzAktualisiert(final DatensatzUpdateEvent event) {
			if (!event.getDatum().isValid()) {
				return;
			}

			long zeitstempel;
			FuzzyVariable ist;

			zeitstempel = event.getDatum().getZeitstempel();
			ist = ((FuzzyWert) event.getDatum()).getFuzzyWert();
			if (ist.isValid()) {
				istWerte.put(zeitstempel, ist);
			}
			if (istWerte.size() >= sollWerte.size()) {
				synchronized (this) {
					notify();
				}
			}
		}

	}

	/** Der Name des Wertes der getestet wird. */
	private final String name;

	/** Die Verbindung zur Datenbank. */
	private final Connection connection;

	/** Der Logger der Klasse. */
	private final Logger log;

	/** Der Empfänger der Ist-Werte. */
	private final Empfaenger empfaenger;

	/** Der Datensatz des Fuzzywerts. */
	private final OnlineDatensatz<FuzzyWert> fuzzyWert;

	/** Ordnet die Sollwerte dem Zeitstempel an dem sie gesendet wurden zu. */
	private final SortedMap<Long, FuzzyVariable> sollWerte;

	/** Ordnet die Istwerte dem Zeitstempel an dem sie gesendet wurden zu. */
	private final SortedMap<Long, FuzzyVariable> istWerte;

	/** Der Datensatz des fuzzyfizierende Messwerts. */
	private final OnlineDatensatz<T> messwert;

	/** Der Aspekt mit dem der Messwert gesendet wird. */
	private final Aspect messwertAsp;

	/** Enthält die Testdaten. */
	private final SortedMap<Long, T> testdaten;

	/**
	 * Initialisiert das Objekt.
	 *
	 * @param objekt
	 *            das Systemobjekt an dem der Test durchgeführt wird.
	 * @param name
	 *            der Name des Wertes der getestet wird.
	 * @param messwert
	 *            der Datensatz des fuzzyfizierende Messwerts.
	 * @param messwertAsp
	 *            der Aspekt der zum Senden der Messwerte verwendet werden soll.
	 * @param fuzzyWert
	 *            der Datensatz des Fuzzywerts.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	public Testfall(final SystemObjekt objekt, final String name,
			final Class<? extends OnlineDatensatz<T>> messwert,
					final Aspect messwertAsp,
					final Class<? extends OnlineDatensatz<FuzzyWert>> fuzzyWert)
					throws SQLException {
		final Configuration konfig;
		final Level level;

		konfig = Configuration.getConfiguration();
		level = Level.parse(konfig.getString("debug.level"));
		log = Logger.getLogger(getClass().getName());
		log.setLevel(level);

		istWerte = new TreeMap<>();
		empfaenger = new Empfaenger();

		this.name = name;
		this.messwert = objekt.getOnlineDatensatz(messwert);
		this.messwertAsp = messwertAsp;
		this.fuzzyWert = objekt.getOnlineDatensatz(fuzzyWert);

		connection = initConnection();
		testdaten = abfragenTestdaten();
		sollWerte = abfragenSoll();
	}

	/**
	 * Initialisiert die Verbindung zur Datenbank.
	 *
	 * @return die offene Datenbankverbindung.
	 */
	private Connection initConnection() {
		Connection conn;

		log.fine(name + ": Lade Datenbanktreiber ...");
		try {
			Class.forName("org.hsqldb.jdbcDriver");
		} catch (final ClassNotFoundException ex) {
			throw new IllegalStateException(
					"Datenbanktreiber nicht gefunden: " + ex);
		}

		log.fine(name + ": Verbinde mit Datenbank ...");
		try {
			conn = DriverManager.getConnection(
					"jdbc:hsqldb:res:de/bsvrz/iav/subsegana/subsegana/prueffall7;shutdown=true",
					"sa", "");
		} catch (final SQLException ex) {
			throw new IllegalStateException(
					"Kann keine Verbindung zur Datenbank aufbauen: " + ex);
		}

		return conn;
	}

	/**
	 * Gibt die Liste der zu sendenden Testdaten zurück.
	 *
	 * @return die Testdaten.
	 */
	public SortedMap<Long, T> getTestdaten() {
		return testdaten;
	}

	/**
	 * Gibt die Liste der Sollwerte zurück.
	 *
	 * @return die Sollwerte.
	 */
	public SortedMap<Long, FuzzyVariable> getSollWerte() {
		return sollWerte;
	}

	/**
	 * Gibt die Liste der Istwerte zurück.
	 *
	 * @return die Istwerte.
	 */
	public SortedMap<Long, FuzzyVariable> getIstWerte() {
		return istWerte;
	}

	/**
	 * Führt den Test aus.
	 */
	public void ausfuehren() {
		int laenge;

		laenge = getTestdaten().size();
		if (laenge < getSollWerte().size()) {
			fail("Die Liste der Testdaten muss mindestens so lang sein wie die der Sollwerte.");
		}

		log.fine(name + ": Anmelden als Empfänger der Ist-Werte ...");
		fuzzyWert.addUpdateListener(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(),
				empfaenger);

		// Testdaten senden.
		log.fine(name + ": Anmelden als Sender (Quelle) der Testdaten ...");
		messwert.setQuelle(messwertAsp, true);
		try {
			messwert.anmeldenSender(messwertAsp);
		} catch (final AnmeldeException ex) {
			ex.printStackTrace();
			fail("Konnte sich nicht als Sender für Testdaten anmelden: " + ex);
		}
		log.fine(name + ": Sende Testdaten ...");
		for (final long t : testdaten.keySet()) {
			final T datum;

			datum = getTestdaten().get(t);
			datum.setZeitstempel(t);
			log.finer("Sende Testdaten für den Zeitpunkt "
					+ com.bitctrl.util.Timestamp.absoluteTime(t) + " (" + t
					+ ") ...");
			try {
				messwert.sendeDaten(messwertAsp, datum);
			} catch (final DatensendeException ex) {
				ex.printStackTrace();
				fail("Konnte Testdaten nicht senden: " + ex);
			}
		}

		// Warten bis alle Antworten empfangen wurden.
		log.fine(name + ": Warte auf Eingang aller Ist-Werte ...");
		if (istWerte.size() < sollWerte.size()) {
			try {
				synchronized (empfaenger) {
					empfaenger.wait(Constants.MILLIS_PER_MINUTE);
				}
			} catch (final InterruptedException ex) {
				ex.printStackTrace();
			}
		}
	}

	/**
	 * Vergleicht die empfangen Ist-Werte mit den Soll-Werten.
	 */
	public void auswerten() {
		log.fine(name
				+ ": Vom Datenverteiler als Sender und Empfänger abmelden ...");
		fuzzyWert.removeUpdateListener(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(),
				empfaenger);
		messwert.abmeldenSender(messwertAsp);

		log.fine(name + ": Vergleiche Soll- und Ist-Werte ...");
		for (final Long t : getSollWerte().keySet()) {
			final String zeitpunkt;
			final FuzzyVariable soll, ist;

			zeitpunkt = Timestamp.absoluteTime(t);
			log.finer("Vergleiche Soll- und Istwert zum Zeitpunkt " + zeitpunkt
					+ " (" + t + ") ...");

			soll = getSollWerte().get(t);
			ist = getIstWerte().get(t);
			assertEquals("Soll und Ist stimmen nicht überein: zeitpunkt="
					+ zeitpunkt, soll, ist);
		}
	}

	/**
	 * Gibt den Datensatz mit dem Messwert zurück.
	 *
	 * @return der Messwertdatensatz.
	 */
	public OnlineDatensatz<T> getMesswert() {
		return messwert;
	}

	/**
	 * Gibt den Aspekt des Messwertdatensatzes zurück.
	 *
	 * @return der Aspekt.
	 */
	public Aspect getMesswertAsp() {
		return messwertAsp;
	}

	/**
	 * Gibt den Datensatz mit dem Fuzzy-Wert zurück.
	 *
	 * @return der Fuzzy-Wertdatensatz.
	 */
	public OnlineDatensatz<FuzzyWert> getFuzzyWert() {
		return fuzzyWert;
	}

	/**
	 * Gibt den Namen des Testfalls zurück.
	 *
	 * @return der Name des Testfalls.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Ruft die Sollwerte aus der Datenbank ab.
	 *
	 * @return die Sollwerte.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	protected abstract SortedMap<Long, FuzzyVariable> abfragenSoll()
			throws SQLException;

	/**
	 * Ruft die Testdaten aus der Datenbank ab.
	 *
	 * @return die Testdaten.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	protected abstract SortedMap<Long, T> abfragenTestdaten()
			throws SQLException;

	/**
	 * Gibt die Verbindung zur Datenbank zurück.
	 *
	 * @return die Datenbankverbindung.
	 */
	public Connection getConnection() {
		return connection;
	}

	/**
	 * Gibt den Logger der Klasse zurück.
	 *
	 * @return der Logger.
	 */
	public Logger getLog() {
		return log;
	}

}
